@@ -51,7 +51,7 @@ else:
 
         def within_tol(x, y, atol, rtol):
             with np.errstate(invalid='ignore'):
-                result = np.less_equal(abs(x-y), atol + rtol * abs(y))
+                result = np.less_equal(abs(x - y), atol + rtol * abs(y))
             if np.isscalar(a) and np.isscalar(b):
                 result = bool(result)
             return result
@@ -93,7 +93,8 @@ else:
 
 
 def _tree_to_labels(X, single_linkage_tree, min_cluster_size=10,
-                    allow_single_cluster=False):
+                    allow_single_cluster=False,
+                    match_reference_implementation=False):
     """ Converts a pretrained tree and cluster size into a
     set of labels and probabilities.
     """
@@ -102,7 +103,8 @@ def _tree_to_labels(X, single_linkage_tree, min_cluster_size=10,
     stability_dict = compute_stability(condensed_tree)
     labels, probabilities, stabilities = get_clusters(condensed_tree,
                                                       stability_dict,
-                                                      allow_single_cluster)
+                                                      allow_single_cluster,
+                                                      match_reference_implementation)
 
     return (labels, probabilities, stabilities, condensed_tree,
             single_linkage_tree)
@@ -148,7 +150,7 @@ def _hdbscan_generic(X, min_samples=5, alpha=1.0, metric='minkowski', p=2,
 
     # Sort edges of the min_spanning_tree by weight
     min_spanning_tree = min_spanning_tree[np.argsort(min_spanning_tree.T[2]),
-                                          :]
+                        :]
 
     # Convert edge list into standard hierarchical clustering format
     single_linkage_tree = label(min_spanning_tree)
@@ -159,7 +161,6 @@ def _hdbscan_generic(X, min_samples=5, alpha=1.0, metric='minkowski', p=2,
 def _hdbscan_sparse_distance_matrix(X, min_samples=5, alpha=1.0,
                                     metric='minkowski', p=2, leaf_size=40,
                                     gen_min_span_tree=False, **kwargs):
-
     assert issparse(X)
 
     lil_matrix = X.tolil()
@@ -171,7 +172,10 @@ def _hdbscan_sparse_distance_matrix(X, min_samples=5, alpha=1.0,
     if csgraph.connected_components(mutual_reachability_, directed=False,
                                     return_labels=False) > 1:
         raise ValueError('Sparse distance matrix has multiple connected'
-                         ' components!\nRun hdbscan on each component.')
+                         ' components!\nThat is, there exist groups of points '
+                         'that are completely disjoint -- there are no distance '
+                         'relations connecting them\n'
+                         'Run hdbscan on each component.')
 
     # Compute the minimum spanning tree for the sparse graph
     sparse_min_spanning_tree = csgraph.minimum_spanning_tree(
@@ -184,7 +188,7 @@ def _hdbscan_sparse_distance_matrix(X, min_samples=5, alpha=1.0,
 
     # Sort edges of the min_spanning_tree by weight
     min_spanning_tree = min_spanning_tree[np.argsort(min_spanning_tree.T[2]),
-                                          :][0]
+                        :][0]
 
     # Convert edge list into standard hierarchical clustering format
     single_linkage_tree = label(min_spanning_tree)
@@ -198,6 +202,8 @@ def _hdbscan_sparse_distance_matrix(X, min_samples=5, alpha=1.0,
 def _hdbscan_prims_kdtree(X, min_samples=5, alpha=1.0,
                           metric='minkowski', p=2, leaf_size=40,
                           gen_min_span_tree=False, **kwargs):
+    if X.dtype != np.float64:
+        X = X.astype(np.float64)
 
     # The Cython routines used require contiguous arrays
     if not X.flags['C_CONTIGUOUS']:
@@ -218,7 +224,7 @@ def _hdbscan_prims_kdtree(X, min_samples=5, alpha=1.0,
 
     # Sort edges of the min_spanning_tree by weight
     min_spanning_tree = min_spanning_tree[np.argsort(min_spanning_tree.T[2]),
-                                          :]
+                        :]
 
     # Convert edge list into standard hierarchical clustering format
     single_linkage_tree = label(min_spanning_tree)
@@ -229,6 +235,8 @@ def _hdbscan_prims_kdtree(X, min_samples=5, alpha=1.0,
 def _hdbscan_prims_balltree(X, min_samples=5, alpha=1.0,
                             metric='minkowski', p=2, leaf_size=40,
                             gen_min_span_tree=False, **kwargs):
+    if X.dtype != np.float64:
+        X = X.astype(np.float64)
 
     # The Cython routines used require contiguous arrays
     if not X.flags['C_CONTIGUOUS']:
@@ -248,7 +256,7 @@ def _hdbscan_prims_balltree(X, min_samples=5, alpha=1.0,
                                                 alpha)
     # Sort edges of the min_spanning_tree by weight
     min_spanning_tree = min_spanning_tree[np.argsort(min_spanning_tree.T[2]),
-                                          :]
+                        :]
     # Convert edge list into standard hierarchical clustering format
     single_linkage_tree = label(min_spanning_tree)
 
@@ -260,13 +268,15 @@ def _hdbscan_boruvka_kdtree(X, min_samples=5, alpha=1.0,
                             approx_min_span_tree=True,
                             gen_min_span_tree=False,
                             core_dist_n_jobs=4, **kwargs):
-
     if leaf_size < 3:
         leaf_size = 3
 
     if core_dist_n_jobs < 1:
         core_dist_n_jobs = max(cpu_count() + 1 + core_dist_n_jobs, 1)
 
+    if X.dtype != np.float64:
+        X = X.astype(np.float64)
+
     tree = KDTree(X, metric=metric, leaf_size=leaf_size, **kwargs)
     alg = KDTreeBoruvkaAlgorithm(tree, min_samples, metric=metric,
                                  leaf_size=leaf_size // 3,
@@ -290,13 +300,15 @@ def _hdbscan_boruvka_balltree(X, min_samples=5, alpha=1.0,
                               approx_min_span_tree=True,
                               gen_min_span_tree=False,
                               core_dist_n_jobs=4, **kwargs):
-
     if leaf_size < 3:
         leaf_size = 3
 
     if core_dist_n_jobs < 1:
         core_dist_n_jobs = max(cpu_count() + 1 + core_dist_n_jobs, 1)
 
+    if X.dtype != np.float64:
+        X = X.astype(np.float64)
+
     tree = BallTree(X, metric=metric, leaf_size=leaf_size, **kwargs)
     alg = BallTreeBoruvkaAlgorithm(tree, min_samples, metric=metric,
                                    leaf_size=leaf_size // 3,
@@ -305,7 +317,7 @@ def _hdbscan_boruvka_balltree(X, min_samples=5, alpha=1.0,
     min_spanning_tree = alg.spanning_tree()
     # Sort edges of the min_spanning_tree by weight
     min_spanning_tree = min_spanning_tree[np.argsort(min_spanning_tree.T[2]),
-                                          :]
+                        :]
     # Convert edge list into standard hierarchical clustering format
     single_linkage_tree = label(min_spanning_tree)
 
@@ -319,8 +331,8 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
             metric='minkowski', p=2, leaf_size=40,
             algorithm='best', memory=Memory(cachedir=None, verbose=0),
             approx_min_span_tree=True, gen_min_span_tree=False,
-            core_dist_n_jobs=4, allow_single_cluster=False, **kwargs):
-
+            core_dist_n_jobs=4, allow_single_cluster=False,
+            match_reference_implementation=False, **kwargs):
     """Perform HDBSCAN clustering from a vector array or distance matrix.
 
     Parameters
@@ -399,6 +411,13 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
         the case that you feel this is a valid result for your dataset.
         (default False)
 
+    match_reference_implementation : bool, optional (default=False)
+        There exist some interpretational differences between this
+        HDBSCAN* implementation and the original authors reference
+        implementation in Java. This can result in very minor differences
+        in clustering results. Setting this flag to True will, at a some
+        performance cost, ensure that the clustering results match the
+        reference implementation.
 
     **kwargs : optional
         Arguments passed to the distance metric
@@ -471,6 +490,11 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
             raise ValueError('Minkowski metric with negative p value is not'
                              ' defined!')
 
+    if match_reference_implementation:
+        min_samples = min_samples - 1
+        min_cluster_size = min_cluster_size + 1
+        approx_min_span_tree = False
+
     # Checks input and converts to an nd-array where possible
     X = check_array(X, accept_sparse='csr')
     # Python 2 and 3 compliant string_type checking
@@ -486,8 +510,8 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
         if algorithm == 'generic':
             (single_linkage_tree,
              result_min_span_tree) = memory.cache(
-                 _hdbscan_generic)(X, min_samples, alpha, metric,
-                                   p, leaf_size, gen_min_span_tree, **kwargs)
+                _hdbscan_generic)(X, min_samples, alpha, metric,
+                                  p, leaf_size, gen_min_span_tree, **kwargs)
         elif algorithm == 'prims_kdtree':
             if metric not in KDTree.valid_metrics:
                 raise ValueError("Cannot use Prim's with KDTree for this"
@@ -568,11 +592,13 @@ def hdbscan(X, min_cluster_size=5, min_samples=None, alpha=1.0,
     return _tree_to_labels(X,
                            single_linkage_tree,
                            min_cluster_size,
-                           allow_single_cluster) + (result_min_span_tree,)
+                           allow_single_cluster,
+                           match_reference_implementation) + \
+            (result_min_span_tree,)
 
 
 # Inherits from sklearn
-class HDBSCAN (BaseEstimator, ClusterMixin):
+class HDBSCAN(BaseEstimator, ClusterMixin):
     """Perform HDBSCAN clustering from vector array or distance matrix.
 
     HDBSCAN - Hierarchical Density-Based Spatial Clustering of Applications
@@ -649,9 +675,17 @@ class HDBSCAN (BaseEstimator, ClusterMixin):
 
     allow_single_cluster : bool, optional (default=False)
         By default HDBSCAN* will not produce a single cluster, setting this
-        to t=True will override this and allow single cluster results in
+        to True will override this and allow single cluster results in
         the case that you feel this is a valid result for your dataset.
 
+    match_reference_implementation : bool, optional (default=False)
+        There exist some interpretational differences between this
+        HDBSCAN* implementation and the original authors reference
+        implementation in Java. This can result in very minor differences
+        in clustering results. Setting this flag to True will, at a some
+        performance cost, ensure that the clustering results match the
+        reference implementation.
+
     **kwargs : optional
         Arguments passed to the distance metric
 
@@ -719,7 +753,8 @@ class HDBSCAN (BaseEstimator, ClusterMixin):
                  approx_min_span_tree=True,
                  gen_min_span_tree=False,
                  core_dist_n_jobs=4,
-                 allow_single_cluster=False, **kwargs):
+                 allow_single_cluster=False,
+                 match_reference_implementation=False, **kwargs):
         self.min_cluster_size = min_cluster_size
         self.min_samples = min_samples
         self.alpha = alpha
@@ -733,6 +768,7 @@ class HDBSCAN (BaseEstimator, ClusterMixin):
         self.gen_min_span_tree = gen_min_span_tree
         self.core_dist_n_jobs = core_dist_n_jobs
         self.allow_single_cluster = allow_single_cluster
+        self.match_reference_implementation = match_reference_implementation
 
         self._metric_kwargs = kwargs
 
